from RCC8 import *
from copy import deepcopy

class NeighborhoodGraphNode:
    def __init__( self, about ):
        self.about = set( about )
        self.closure = None
        self.interior = None
        self.isPartOf = []
        self.parts = []

    def isClosure( self ):
        return self is self.closure

    def isInterior( self ):
        return self is self.interior

    def isAbout( self, region ):
        return region <= self.about

    def setIsPartOf( self, otherNode ):
        if otherNode not in self.isPartOf:
            self.isPartOf.append( otherNode )
            otherNode.parts.append( self )

    def unsetIsPartOf( self, otherNode ):
        self.isPartOf.remove( otherNode )
        otherNode.parts.remove( self )

    def hasPart( self, otherNode ):
        if self is otherNode:
            return False
        if otherNode in self.parts:
            return True
        for p in self.parts:
            if p.hasPart( otherNode ):
                return True
        return False

    def isChildOf( self, otherNode ):
        return otherNode.hasPart( self )

class Profile:
    def __init__( self ):
        self.active = True
        self.numberOfExploredStates = 0
        self.numberOfQuickFixes = 0
        self.transformationCount = dict(  )

    def increment( self ):
        if self.active:
            self.numberOfExploredStates += 1

    def incrementiQF( self ):
        if self.active:
            self.numberOfQuickFixes += 1

    def count( self, transformation ):
        self.increment(  )
        if self.active:
            self.transformationCount[transformation] = self.transformationCount.get( transformation, 0 ) + 1

class Logger:
    def __init__( self ):
        self.logLines = []

    def __call__( self, *param ):
        line = " ".join( map( str, param ) )
        self.logLines.append( line )
     
class NeighborhoodGraph:
    def __init__( self ):
        self.universe = NeighborhoodGraphNode( ('universe',) )
        self.universe.closure  = self.universe
        self.universe.interior = self.universe
        self.all_regions = [self.universe]
        self.profile = Profile(  )
        self.log = Logger( )

    def createCopy( self ):
        nhg = deepcopy( self )
        nhg.profile = self.profile
        nhg.log = self.log
        return nhg


    def __iter__( self ):
        for v in self.all_regions:
            yield v

    def createNewRegion( self, name, createInterior=True, register=True, partOfUniverse=True, checkNotExist=True ):
        try:
            name = set( (name,) )
        except:
            name = set( name )

        if checkNotExist:
            for region in self.all_regions:
                assert not region.isAbout( name ), ( name, region.about )

        class ClosureNode( NeighborhoodGraphNode ):
            def __init__( self, about ):
                NeighborhoodGraphNode.__init__( self, about )
                self.closure = self

            def createInteriorNode( self ):
                assert self.interior is None

                class InteriorNode( NeighborhoodGraphNode ):
                    def __init__( self, clNode ):
                        NeighborhoodGraphNode.__init__( self, [] )
                        self.closure = clNode
                        self.interior = self
                        self.setIsPartOf( clNode )
                        self.about = clNode.about

                self.interior = InteriorNode( self )

                return self.interior

        cl = ClosureNode( set( name ) )
        if register:
            self.all_regions.append( cl )

        if createInterior:
            i  = cl.createInteriorNode(  )
            if register:
                self.all_regions.append( i )

        if partOfUniverse:
            cl.closure.setIsPartOf( self.universe )

        return cl

    def equal( self, nhg ):
        about = map( lambda r: r.about, self.all_regions )
        nhgAbout = map( lambda r: r.about, nhg.all_regions )

        if len( about ) != len( nhgAbout ):
            return False

        for a in about:
            if a not in nhgAbout:
                return False
        return True

    def removeNode( self, node ):
        for p in node.isPartOf[:]:
            node.unsetIsPartOf( p )
        for p in node.parts[:]:
            p.unsetIsPartOf( node )
        self.all_regions.remove( node )


    def transformDCToEC( self, region1, region2 ):
        self.profile.count( "DCtoEC" )
        region1 = set( region1 )
        region2 = set( region2 )
        assert rcc8( self, region1, region2 ) == dc, "Not disconnected"

        newRegions = list(  )

        #assert len( region1 ) == 1, region1
        #assert len( region2 ) == 1, region2

        for n1 in list(self.getRegions( region1 )):
            for n2 in list(self.getRegions( region2 )):
                p1 = set( n1.isPartOf )
                p2 = set( n2.isPartOf )
                parent = p1 & p2
                if len( parent ) != 1:
                    continue

                parent = list(parent)[0]

                self.log( "Select parent: ", parent.about, "for", region1, "=", n1.about, "and", region2, "=", n2.about )

                #if not( n1 in parent.parts and n2 in parent.parts ):
                #    continue

                # create the new node 
                cl = self.createNewRegion( n1.about | n2.about, False, False, False )
                cl.closure.setIsPartOf( n1.closure )
                cl.closure.setIsPartOf( n2.closure )
                newRegions.append( cl )
                break
            else:
                continue
            break

        assert len( newRegions ) > 0, "Could not create regions"

        self.all_regions += newRegions

        self.quickFixEdges(  )

        assert self.valid(  ), "Not Valid anymore"
        assert rcc8( self, region1, region2 ) == ec, "Not externaly connected afterwards " + rcc8( self, region1, region2 ).__name__
        
    
    def transformECToPO( self, region1, region2 ):
        self.profile.count( "ECtoPO" )
        region1 = set( region1 )
        region2 = set( region2 )
        assert rcc8( self, region1, region2 ) == ec

        cl = self.getClosure( region1 | region2 )
        for p in cl.isPartOf:
            assert p.interior is not None

        i = cl.createInteriorNode(  )
        self.all_regions.append( i )

        for p in cl.isPartOf:
            i.setIsPartOf( p.interior )
        assert rcc8( self, region1, region2 ) == po
        assert self.valid(  )

    def transformPOtoTPP( self, region1, region2 ):
        self.profile.count( "POtoTPP" )
        region1 = set( region1 )
        region2 = set( region2 )
        assert rcc8( self, region1, region2 ) == po

        r1 = self.getClosure( region1 )
        r  = self.getClosure( region1 | region2 )

        o = None
        for p in r1.closure.isPartOf:
            if p.isInterior(  ):
                assert o is None, str( r1.about ) + " is in more then one open region"
                o = p

        for pp_region in r1.interior.parts:
            assert not pp_region.isClosure(  )

        for cl_region in list(self.getClosures( region1 )):
            if not cl_region.isAbout( region2 ):
                self.removeNode( cl_region )
                if cl_region.interior is not None:
                    self.removeNode( cl_region.interior )

        r.closure.setIsPartOf( o )

        self.quickFixEdges(  )

        assert self.valid(  )
        assert rcc8( self, region1, region2 ) == tpp

    def transformTPPtoNTPP( self, region1, region2 ):
        self.profile.count( "TPPtoNTPP" )
        region1 = set( region1 )
        region2 = set( region2 )
        assert rcc8( self, region1, region2 ) == tpp


        r1 = self.getClosure( region1 )

        r2c = filter( lambda r: r in r1.isPartOf, self.getClosures( region2 ) )

        assert len( r2c ) == 1
        r2 = r2c[0]

        for p in r1.closure.parts[:]:
            if p.isClosure(  ) and not r1.interior.hasPart( p.interior )  and not r2.interior.hasPart( p.interior ):
                self.removeNode( p )
                if p.interior is not None:
                    self.removeNode( p.interior )

        for p in r1.closure.isPartOf:
            if not p.isAbout( r2.about ):
                r1.unsetIsPartOf( p )

        r1.unsetIsPartOf( r2 )
        r1.interior.unsetIsPartOf( r2.interior )
        r1.setIsPartOf( r2.interior )

        self.quickFixEdges(  )
                    
        assert self.valid(  )
        assert rcc8( self, region1, region2 ) == ntpp

    def transformTPPitoNTPPi( self, region1, region2 ):
        self.transformTPPtoNTPP( region2, region1 )

    def transformPOtoTPPi( self, region1, region2 ):
        self.transformPOtoTPP( region2, region1 )

    def transformPOtoEQ( self, region1, region2 ):
        self.profile.count( "POtoEQ" )
        region1 = set( region1 )
        region2 = set( region2 )

        assert rcc8( self, region1, region2 ) == po, "Regions are not po but are " + rcc8( self, region1, region2 ).__name___

        r1 = self.getClosure( region1 )
        r2 = self.getClosure( region2 )
        r  = self.getClosure( region1 | region2 ) # region1 and region2
        p  = self.getClosure( region1 & region2 ) # parent of region1 and region2

        # ensure that nothing is only connected to region1 or region2
        for region in self.getRegions( region1 ):
            assert region is r1.closure or region is r1.interior or region.isAbout( region2 ), str( region.about ) + " " + str( region.isClosure(  ) ) + " is part of r1 but not in the overlap"
        for region in self.getRegions( region2 ):
            assert region is r2.closure or region is r2.interior or region.isAbout( region1 ), str( region.about ) + " " + str( region.isClosure(  ) ) + " is part of r2 but not in the overlap"


        self.removeNode( r1 )
        self.removeNode( r1.interior )

        self.removeNode( r2 )
        self.removeNode( r2.interior )

        r.setIsPartOf( p )

        assert rcc8( self, region1, region2 ) == eq, "They are not equal afterwards but are" + rcc8( self, region1, region2 ).__name__
        assert self.valid(  ), "The NGH is not valid anymore"

    def transformEQtoNTPP( self, region1, region2 ):
        self.profile.count( "EQtoNTPP" )
        region1 = set( region1 )
        region2 = set( region2 )

        assert rcc8( self, region1, region2 ) == eq, "Not equal at start"

        r2 = self.getClosure( region2 )
        r2.closure.about  -= region1
        r2.interior.about -= region1

        connect = list(  )

        for cl in list(self.getClosures( region1 )):
            if r2.interior.hasPart( cl ):
                continue
            if cl.interior is not None:
                connect.append( cl )

            cl.closure.about  -= region1
            cl.interior.about -= region1

        r1 = self.createNewRegion( region1 | r2.about, True, True, False, False )

        for p in r2.interior.parts[:]:
            if p.isAbout( region1 ):
                p.unsetIsPartOf( r2.interior )
                p.setIsPartOf( r1.interior )

        r1.closure.setIsPartOf( r2.interior )

        for r in connect:
            n = self.createNewRegion( region1 | r2.about | r.about, True, True, False, True )
            n.closure.setIsPartOf( r1.closure )
            n.closure.setIsPartOf( r.closure )
            n.interior.setIsPartOf( r1.interior )
            n.interior.setIsPartOf( r.interior )

        assert self.valid(  ), "Not valid anymore"
        assert rcc8( self, region1, region2 ) == ntpp, "Not ntpp afterwards"

    def transformEQtoNTPPi( self, region1, region2 ):
        self.transformEQtoNTPP( region2, region1 )

    def transformTPPtoEQ( self, region1, region2 ):
        self.profile.count( "TPPtoEQ" )
        region1 = set( region1 )
        region2 = set( region2 )

        assert rcc8( self, region1, region2 ) == tpp

        r1 = self.getClosure( region1 )
        r2 = self.getClosure( region2 )

        assert len( r1.closure.isPartOf ) == 2, map(lambda r: r.about, r1.closure.isPartOf )
        assert len( r2.interior.parts ) == 1

        r1.closure.unsetIsPartOf( r2.closure )
        r1.interior.unsetIsPartOf( r2.interior )

        for p in r2.closure.isPartOf[:]:
            r2.closure.unsetIsPartOf( p )
            r1.closure.setIsPartOf( p )

        self.all_regions.remove( r2.closure )
        self.all_regions.remove( r2.interior )

        assert rcc8( self, region1, region2 ) == eq
        assert self.valid(  )

    def transformTPPitoEQ( self, region1, region2 ):
        self.transformTPPtoEQ( region2, region1 )

    def toDot( self ):
        """Return the graph in the dot fileformat as a string"""
        out  = "digraph G{\n"
        out += "\tnode [shape=plaintext];\n"
        out += "\tedge [dir=back];\n"
        out += "\n"
        out += self.__toDotNode__( self.universe, set(  ) )
        out += "}"
        return out

    def __toDotEdge__( self, parent, child ):
        out = "\t"
        out += self.__toDotName__( parent )
        out += " -> "
        out += self.__toDotName__( child );
        out += ";\n"
        return out


    def __toDotName__( self, node ):
        out = ""
        if node.isClosure(  ):
            out += "CL_"
        out +="_".join( map( str, sorted( node.about ) ) )
        return out


    def __toDotNode__( self, node, nodeWritten ):
        out = ""
        if node not in nodeWritten:
            out += "\t"
            out += self.__toDotName__( node )
            out += ";\n"
            nodeWritten.add( node )

            for child in node.parts:
                out += self.__toDotNode__( child, nodeWritten )
                out += self.__toDotEdge__( node, child )

        return out

    def __str__( self ):
        return self.toDot(  )

    def valid( self ):
        for region in self.all_regions:
            if region.about is None or len( region.about ) == 0:
                return False

            for p in region.parts:
                if region not in p.isPartOf:
                    self.log( "NOT VALID 1:", region, "not in", p.isPartOf )
                    self.log( self.toDot(  ) )
                    return False
                if p not in self.all_regions:
                    self.log( "NOT VALID 1:" )
                    self.log( region.about, region.isClosure(  ) )
                    self.log( (p.about,p.isClosure( ) ), "is not in", map(lambda r: ( r.about, r.isClosure(  ) ), self.all_regions ) )
                    self.log( self.toDot(  ) )
                    return False
                if not p.isAbout( region.about ) and region is not self.universe:
                    self.log( "NOT VALID 1:", region.about, p.about  )
                    self.log( self.toDot(  ) )
                    return False
            for p in region.isPartOf:
                if region not in p.parts:
                    self.log( "NOT VALID 2:", region, "not in", p.parts )
                    self.log( self.toDot(  ) )
                    return False
                if p not in self.all_regions:
                    self.log( "NOT VALID 2:" )
                    self.log( region.about, region.isClosure(  ) )
                    self.log( (p.about,p.isClosure( ) ), "is not in", map(lambda r: ( r.about, r.isClosure(  ) ), self.all_regions ) )
                    self.log( self.toDot(  ) )
                    return False
                if not region.isAbout( p.about ) and p is not self.universe:
                    self.log( "NOT VALID 2:", region.about, p.about  )
                    self.log( self.toDot(  ) )
                    return False

        for r_name in self.getRegionNames(  ):
            r = self.getClosure( set([r_name]) )
            if r is None:
                self.log( r_name, "not found" )
                return False
            elif r.interior is None:
                self.log( r_name, "no interior found" )
                return False

        for r1 in self.getRegionNames(  ):
            for r2 in self.getRegionNames(  ):
                try:
                    rcc8( self, set( [r1] ), set( [r2]) )
                except:
                    self.log( r1, r2, "no rcc8 relation found" )
                    return False

        return True

    def quickFixEdges( self ):
        for region in self.all_regions:
            for p in self.all_regions:
                if p is region:
                    continue
                if p.isClosure(  ) == region.isClosure(  ) and region.isAbout( p.about ):  # closed and region should be below and conntect to p
                    if not p.hasPart( region ):
                        self.log( "Quick Fix:", region.about, "connect to", p.about )
                        region.setIsPartOf( p )


    def getClosure( self, about ):
        about = about - self.universe.about

        if len( about ) == 0:
            return self.universe

        cl = list(self.getClosures( about ))
        tmp = list(  )
        for r1 in cl:
            for r2 in cl:
                if r1 is r2:
                    continue
                if not r1.hasPart( r2 ):
                    break
            else:
                tmp.append( r1 )

        if len( tmp ) == 0:
            return None

        if len( tmp ) == 1:
            return tmp[0]

        raise Exception( "Undefined Closure", about, map(lambda r: r.about, tmp) )

    def getClosures( self, about ):
        for region in self.all_regions:
            if region.isClosure( ) and region.isAbout( about ):
                yield region

    def getRegions( self, about ):
        for region in self.all_regions:
            if region.isAbout( about ) :
                yield region

    def getRegionNames( self ):
        """Return all 'simple' region names''"""
        names = set(  )
        for region in self.all_regions:
            for name in region.about:
                names.add( name )
        names -= self.universe.about
        return names

    def getPossibleTransitions( self, alsoUp = False, regions = None, checkPossible=True, constrains = None ):
        base_transitions = dict(  )
        base_transitions[dc]    = [NeighborhoodGraph.transformDCToEC]
        base_transitions[ec]    = [NeighborhoodGraph.transformECToPO] 
        base_transitions[po]    = [NeighborhoodGraph.transformPOtoTPP,  NeighborhoodGraph.transformPOtoTPPi, NeighborhoodGraph.transformPOtoEQ] 
        base_transitions[eq]    = [NeighborhoodGraph.transformEQtoNTPP, NeighborhoodGraph.transformEQtoNTPPi] 
        base_transitions[tpp]   = [NeighborhoodGraph.transformTPPtoEQ,  NeighborhoodGraph.transformTPPtoNTPP] 
        base_transitions[tppI]  = [NeighborhoodGraph.transformTPPitoEQ, NeighborhoodGraph.transformTPPitoNTPPi] 
        base_transitions[ntpp]  = []
        base_transitions[ntppI] = []

        if alsoUp:
            base_transitions[ec]    += [NeighborhoodGraph.transformECToDC]
            base_transitions[po]    += [NeighborhoodGraph.transformPOtoEC]
            base_transitions[eq]    += [NeighborhoodGraph.transformEQtoTPP,    NeighborhoodGraph.transformEQtoTPPi,NeighborhoodGraph.transformEQtoPO]
            base_transitions[tpp]   += [NeighborhoodGraph.transformTPPtoPO]
            base_transitions[tppI]  += [NeighborhoodGraph.transformTPPitoPO]
            base_transitions[ntpp]  += [NeighborhoodGraph.transformNTPPtoTPP,  NeighborhoodGraph.transformNTPPtoEQ]
            base_transitions[ntppI] += [NeighborhoodGraph.transformNTPPitoTPP, NeighborhoodGraph.transformNTPPitoEQ]

        transitions = list(  )

        self.log( regions, "=>", )

        if regions is None:
            regions = set(  )
            for r in self.all_regions:
                if r.interior is not None and r is not self.universe and len( r.about ) > 1:
                    regions.add( tuple(r.about) )
            for name in self.getRegionNames(  ):
                f = frozenset( (name,) )
                regions.add( f )
        else:
            tmp = map( frozenset, regions )
            regions = set(  )
            self.log( "TMP:",tmp )
            for t in tmp:
                for r in self.all_regions:
                    if r.interior is not None and ( r.isAbout( t ) ):
                        regions.add( tuple( r.about )  )
                regions.add( t )

        self.log( "Regions:",regions )

        regions =  map(set, regions)

        for r1 in regions:
            for r2 in regions:
                if r1 is r2:
                    continue
                rel = rcc8( self, r1, r2 )
                for trans in base_transitions[rel]:
                    if not checkPossible or self.checkIsPossible( r1, r2, trans, constrains ):
                        transitions.append( ( r1, r2, trans ) )

        return transitions

    def checkIsPossible( self, r1, r2, transition, constrains ):
        nhg = self.createCopy(  )
        nhg.profile.active = False
	try:

		try:
		    transition( nhg, r1, r2 )
		    if nhg.valid(  ):
			return nhg.checkConstrains( constrains )
		finally:
		    self.profile.active = True

        except Exception, e:
            self.log( "Check for", r1, r2, transition.__name__,"failed because", e )
            return False

    def findLowestAncestor( self, r1, r2, node = None ):
        if node is None:
            node = self.universe

        for p in node.parts:
            if p.hasPart( r1 ) and p.hasPart( r2 ):
                return self.findLowestAncestor( r1, r2, p )

        return node


    def transformTo( self, r1, r2, rel, constrains = None, alsoUpwards = False, gready = False, knownStates = None  ):
        r1 = set( r1 )
        r2 = set( r2 )
        self.log( "Call: ", r1, r2, rel.__name__ )

        if rcc8( self, r1, r2 ) == rel:
            yield ( [], [] )
            return

        if knownStates is None:
            knownStates = list(  )

        # skip known states
        if reduce(lambda x,y: x or y, map( lambda s: s.equal( self ), knownStates ), False ):
            yield None
            return

        knownStates.append( self )

        if constrains is None:
            constrains = list(  )

        for constrain in constrains:
            if not constrain( self ):
                self.log( "Constrain violated: ", constrain.__doc__ )
                yield None
                return

        transitions = self.getPossibleTransitions( alsoUpwards, [r1, r2], constrains=constrains )
        if len( transitions ) == 0:
            # no transiton will be possible involving the two regions
            # check if you can still hope
            if len( self.getPossibleTransitions( alsoUpwards, [r1, r2], False ) ) > 0:
                tmp = self.getPossibleTransitions( constrains=constrains )
                self.log( tmp )
                region1 = self.getClosure( r1 )
                region2 = self.getClosure( r2 )
                for t_r1, t_r2, t, in self.getPossibleTransitions( constrains=constrains ):
                    n1 = self.getClosure( t_r1 )
                    n2 = self.getClosure( t_r2 )
                    if lessEqual( region1, n1 ) and lessEqual( region2, n2 ):
                        transitions.append( ( t_r1, t_r2, t ) )
                    if lessEqual( region2, n1 ) and lessEqual( region1, n2 ):
                        transitions.append( ( t_r1, t_r2, t ) )
  
        for transition in transitions:
            nhg = self.createCopy(  )
            t_r1, t_r2, t = transition
            self.log( "transition ...",t_r1,t_r2,t.__name__ )
            t( nhg, t_r1, t_r2 )

            if not nhg.checkConstrains( constrains ):
                continue

            for solution in nhg.transformTo( r1, r2, rel, constrains, alsoUpwards, gready, knownStates ):
                if solution is not None:
                    yield [transition]+solution[0], [nhg]+solution[1]
                    if gready:
                        return
                else:
                    self.log( "else ..." )
                    pass

    def checkConstrains( self, constrains ):
        if constrains is None:
            return True

        good = True
        for c in constrains:
            self.log( c.__doc__, )
            if c( self ):
                self.log( "[OK]" )
            else:
                self.log( "[FAILED]" )
                good = False
        return good

