from RCC8 import *
from Graph import NeighborhoodGraph
from copy import deepcopy
from itertools import permutations
import random
import os
from TimeOut import timeout

def tokenize( line ):
    if line[-1] == '\n':
        line = line[:-1]
    tokens = []
    for t in line.split( " " ):
        if t[0] == '(' and len( t ) > 1:
            tokens.append( t[0] )
            tokens.append( t[1:] )
        elif t[-1] == ')' and len( t ) > 1:
            tokens.append( t[:-1] )
            tokens.append( t[-1] )
        else:
            tokens.append( t )
    return tokens

def createGraphFromFile( filename ):
    baseRel = [ dc, ec, po, tpp, ntpp, eq, tppI, ntppI ]
    relMap = dict( map(lambda r: (str.lower( r.__name__ ),r), baseRel ) )

    buildSteps = []

    f = open( filename )
    for lno, line in enumerate( f ):
        tokens = tokenize( line )
        assert tokens[0] == '(', "Expected '(' but found '" + str( tokens ) + "' in " + line + " @" + str( lno )
        r1 = tokens[1]
        rel = tokens[2]
        r2 = tokens[3]
        assert tokens[4] == ')', "Expected ')' but found '" + str( tokens ) + "' in " + line + " @" + str( lno )

        rel = relMap[rel.lower(  )]

        buildSteps.append( (r1, r2, rel ) )
    f.close(  )

    tmp  = filter( lambda bs: bs[2] is eq,    buildSteps )
    tmp += filter( lambda bs: bs[2] is dc,    buildSteps )
    tmp += filter( lambda bs: bs[2] is ntpp,  buildSteps )
    tmp += filter( lambda bs: bs[2] is ntppI, buildSteps )
    tmp += filter( lambda bs: bs[2] is tpp,   buildSteps )
    tmp += filter( lambda bs: bs[2] is tppI,  buildSteps )
    tmp += filter( lambda bs: bs[2] is po,    buildSteps )
    tmp += filter( lambda bs: bs[2] is ec,    buildSteps )

    assert len( filter( lambda bs: bs not in tmp, buildSteps ) ) == 0

    buildSteps = tmp

    print("-+-"*12)
    print(filename)
    for idx, nhg in enumerate( build( buildSteps ) ):
        return nhg

def profile( folder, num, f ):
    if num is not None:
        selected_filenames = random.sample( folder, num ) 
    else:
        selected_filenames = sorted( folder )
    results = list(  )
    for filename in selected_filenames:
        try:
            nhg = timeout(lambda: createGraphFromFile( filename ), 1200) # timeout in sec
            assert nhg is not None, "No Solution given or timeout"
            p = nhg.profile
            d = ( filename, len( nhg.all_regions ), p.numberOfExploredStates )
            results.append( d )
            f.write( "%s\t%s\t%s\n" % tuple( map( str, d ) ) )
            f.flush(  )
            os.fsync(f.fileno())
        except Exception as e:
            print("ERROR", e.__class__.__name__, e)
            import traceback
            traceback.print_exception( *sys.exc_info(  ) )
            if len( e.args ) > 1:
                try:
                    l = open( filename+".log", "w" )
                    l.write( "\n".join( e.args[1].log.logLines ) )
                    l.close(  )
                except:
                    pass
    return results

def build( buildSteps ):
    createdRegions = set(  )
    constrains = []
    nhg = NeighborhoodGraph(  )
    for BS in permutations( buildSteps ):
        for solution in buildSolutions( nhg, createdRegions, constrains, BS ):
            has_solution = True
            yield solution
            return
    raise Exception( "No Solution", nhg )

def buildSolutions( nhg, createdRegions, constrains, buildSteps):
    #print("-"*6, "Start:")
    #print(createdRegions)
    #print(map(lambda x: x.__doc__, constrains))
    #if len(buildSteps) > 0:
    #    print(buildSteps[0][0], buildSteps[0][1], buildSteps[0][2].__name__,"...",len( buildSteps ))
    #
    #print("Writing partial solution",)
    #filename = "partial_solution_%02d.dot" % len( buildSteps ) 
    #print(filename)
    #f = open( filename, "w" )
    #f.write( nhg.toDot(  ) )
    #f.close(  )

    for c in constrains:
        assert c( nhg ), "Constrain violated: " + c.__doc__

    if len( buildSteps ) == 0:
        yield nhg
        return

    buildStep = buildSteps[0]
    r1, r2, rel = buildStep
    for r in [r1,r2]:
        if r not in createdRegions:
            #print("Creating region ", r)
            nhg.createNewRegion( r )
            createdRegions.add( r )

    for path in nhg.transformTo( [r1], [r2], rel, constrains ):
        new_constrain = Constrain( [r1], rel, [r2] )
        if len( path[0] ) == 0:
            nhg = deepcopy( nhg )
            assert new_constrain( nhg )
            for solution in buildSolutions( nhg, deepcopy(createdRegions), constrains + [new_constrain], buildSteps[1:] ):
                yield solution
        else:
            nhg = deepcopy( path[1][-1] )
            if not new_constrain( nhg ):
                #print(r1, r2, rcc8( nhg, r1, r2 ).__name__, "!=", rel)
                continue
            for solution in buildSolutions( nhg, deepcopy(createdRegions), constrains + [new_constrain], buildSteps[1:] ):
                yield solution
    #print("+"*6,"End")

if __name__ == "__main__":
    import sys

    path = sys.argv[1]
    try:
        num  = int( sys.argv[2] )
    except:
        num = None
    try:
        files = map(lambda fn: os.path.join( path, fn ),  filter( lambda f: f[-3:] == 'csp', os.listdir( path ) ) )
    except:
        files = [path]

    path = path.replace( '/', '_' )
    if num is not None:
        path += "_" + str( num )

    f = open( path+".csv", 'a' )
    p = profile( files, num, f )
    f.close(  )

